/**
 * Copyright 2011-2013 - University of Queensland
 * 
 * Ontology Annotation Services
 * 
 * Released under CC-BY-SA-3.0 License, with attribution using a URL link to
 * https://github.com/ansell/oas
 */

if (typeof oas === undefined) {
    oas = {};
}

if (typeof oas.ontology == 'undefined') {
    oas.ontology = {};
}

/**
 * Creates a new rdfquery.js databank, adds common prefixes to it, and then
 * returns the databank to the caller.
 */
oas.newDatabank = function() {
    var nextDatabank = $.rdf.databank();
    // TODO: Is base useful to us?
    // nextDatabank.base("http://www.example.org/")
    nextDatabank.prefix("dcterms", "http://purl.org/dc/terms/");
    // NOTE: foaf:maker is added automatically based on who is logged into OAS
    nextDatabank.prefix("foaf", "http://xmlns.com/foaf/0.1/");
    nextDatabank.prefix("moat", "http://moat-project.org/ns#");
    nextDatabank.prefix("rdfs", "http://www.w3.org/2000/01/rdf-schema#");
    nextDatabank.prefix("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
    nextDatabank.prefix("tagging", "http://www.holygoat.co.uk/owl/redwood/0.1/tags/");
    nextDatabank.prefix("oas", "http://purl.org/oas/");

    return nextDatabank;
};

/**
 * Submits a new annotation to the server based on the information in the given
 * elements.
 */
oas.defaultDialogSubmit = function(annotationTarget, annotationContent, ontologyLabel, ontologyUri, ontologyTermLabel,
        ontologyTermUri, resultDialog) {
    if (typeof console != "undefined" && console.debug) {
        console.debug("Submitting form");
    }
    // Pull out the relevant fields from the form and serialise them to
    // the annotation schema

    var nextDatabank = oas.newDatabank();

    // Add type statement, attached to a blank node that will be replaced by a
    // URI after submission
    nextDatabank.add("_:next rdf:type tagging:Tagging .");

    var annotationTargetValue = annotationTarget.val();

    if (annotationTargetValue.length === 0) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Annotation Target URI was not found: annotationTarget=");
            console.debug(annotationTarget);
        }
        return false;
    }
    nextDatabank.add("_:next tagging:taggedResource <" + annotationTargetValue + "> .");

    var annotationContentValue = annotationContent.val();
    if (annotationContentValue.length === 0) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Annotation content was not found: annotationContent=");
            console.debug(annotationContent);
        }
        return false;
    }
    nextDatabank.add('_:next tagging:name "' + annotationContentValue + '" .');

    var ontologyTermUriValue = ontologyTermUri.val();
    if (ontologyTermUriValue.length === 0) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Ontology Term URI was not found: ontologyTermUri=");
            console.debug(ontologyTermUri);
        }
        return false;
    }
    nextDatabank.add("_:next moat:tagMeaning <" + ontologyTermUriValue + "> .");

    // -------------------------------------------------
    // OPTIONAL ITEMS FOLLOW
    // -------------------------------------------------
    // add optional mappings detailing the type of the object if it was
    // available and the ontology that the term was defined in if it was
    // available

    var ontologyTermLabelValue = ontologyTermLabel.val();
    if (ontologyTermLabelValue.length === 0) {
        if (typeof console != "undefined" && console.debug) {
            // NOTE: This is not compulsory
            console.debug("Ontology Term Label was not found: ontologyTermLabel=");
            console.debug(ontologyTermLabel);
        }
    }
    else {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Adding rdfs:label relation");
            console.debug(ontologyTermLabelValue);
        }
        nextDatabank.add("<" + ontologyTermUriValue + "> rdfs:label <" + ontologyTermLabelValue + "> .");
    }

    var ontologyUriValue = ontologyUri.val();
    if (ontologyUriValue.length === 0) {
        if (typeof console != "undefined" && console.debug) {
            // NOTE: This is not compulsory
            console.debug("Ontology URI was not found: ontologyUri=");
            console.debug(ontologyUri);
        }
    }
    else {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Adding rdfs:isDefinedBy relation");
            console.debug(ontologyUriValue);
        }
        nextDatabank.add("<" + ontologyTermUriValue + "> rdfs:isDefinedBy <" + ontologyUriValue + "> .");

        // Only attempt to parse the ontology label if there was a URI attached
        // to it
        var ontologyLabelValue = ontologyLabel.val();
        if (ontologyLabelValue.length === 0) {
            if (typeof console != "undefined" && console.debug) {
                // NOTE: This is not compulsory
                console.debug("Ontology Label was not found: ontologyLabel=");
                console.debug(ontologyLabel);
            }
        }
        else {
            if (typeof console != "undefined" && console.debug) {
                console.debug("Adding rdfs:label relation");
                console.debug(ontologyLabelValue);
            }
            nextDatabank.add("<" + ontologyUriValue + "> rdfs:label <" + ontologyLabelValue + "> .");
        }
    }

    // TODO: Support object type in interface
    // nextDatabank.add("annotationTargetValue rdf:type annotationTargetType
    // .");

    if (typeof console != "undefined" && console.debug) {
        console.debug("annotation databank");
        console.debug(nextDatabank);

        console.debug($.toJSON(nextDatabank.dump({
            format : 'application/json'
        })));
    }

    $.ajax({
        type : "POST",
        url : oas.baseUrl + "annotation/create",
        dataType : "json",
        contentType : "application/rdf+json; charset=utf-8",
        data : $.toJSON(nextDatabank.dump({
            format : 'application/json'
        })),
        success : function(data, textStatus, jqXHR) {
            if (typeof console != "undefined" && console.debug) {
                console.debug("successfully posted to annotation creation service");
                console.debug(data);
                console.debug(textStatus);
                console.debug(jqXHR);
            }

            var annotations = oas.parseAnnotationsResults("", data);
            if (typeof console != "undefined" && console.debug) {
                console.debug("found annotations");
                console.debug(annotations);
            }

            if (annotations.length > 0) {
                // Switch to the result dialog with a link to the resulting
                // object URI
                var anchor = $('<a>').attr('href',
                        oas.baseUrl + "annotation/id/?annotationid=" + annotations[0].annotationUri).append(
                        $('<span>').text(annotations[0].label + " by " + annotations[0].author));

                resultDialog.find(".oasAnnotationResult").append(anchor);
                resultDialog.dialog('open');
            }
        },
        error : function(xhr, ajaxOptions, thrownError) {
            // Just reopen the dialog for now if the submission fails
            inputDialog.dialog("open");
        }
    });

    // Return true as the annotation submission process is asynchronous
    return true;
};

/**
 * Parses an RDF/JSON string containing search results and returns the results
 * as an array which can be used with the autocomplete plugin.
 */
oas.parseSearchResults = function(/* string */baseUrl, /* rdf/json */data) {

    if (typeof console != "undefined" && console.debug) {
        console.debug("Parsing search results");
        // console.debug(baseUrl);
        // console.debug(data);
    }
    var nextDatabank = oas.newDatabank();

    var rdfSearchResults = nextDatabank.load(data);
    if (typeof console != "undefined" && console.debug) {
        console.debug("about to debug rdfSearchResults");
        console.debug(rdfSearchResults);
        console.debug("About to create nextQuery");
    }

    var nextQuery = $.rdf({
        databank : nextDatabank
    }).where(' ?result oas:ontologySearchResult ?ontology').optional('?result rdfs:label ?label').optional(
            '?result rdf:type ?type');
    if (typeof console != "undefined" && console.debug) {
        // console.debug("About to debug nextQuery");
        // console.debug(nextQuery);
    }
    var bindings = nextQuery.select();

    if (typeof console != "undefined" && console.debug) {
        console.debug("Found bindings");
        console.debug(bindings);
    }

    // NOTE: make sure that the data here is filtered as specified in:
    // http://api.jqueryui.com/autocomplete/#option-source
    // http://api.jqueryui.com/autocomplete/#event-response
    var nodeChildren = [];
    $.each(bindings, function(index, value) {
        var nextChild = {};
        if (value.label) {
            nextChild.label = value.label.value;
            nextChild.value = value.result.value;
        }
        else {
            nextChild.label = value.result.value;
            nextChild.value = value.result.value;
        }

        if (value.type) {
            nextChild.type = value.type.value;
        }

        nodeChildren.push(nextChild);
    });

    if (typeof console != "undefined" && console.debug) {
        console.debug("About to return");
        console.debug(nodeChildren);
    }

    return nodeChildren;
};

/**
 * Parses an RDF/JSON string containing search results and returns the results
 * as an array which can be used with the autocomplete plugin.
 */
oas.parseAnnotationsResults = function(/* string */baseUrl, /* rdf/json */data) {

    if (typeof console != "undefined" && console.debug) {
        console.debug("Parsing annotation results");
        // console.debug(baseUrl);
        // console.debug(data);
    }
    var nextDatabank = oas.newDatabank();

    var annotationResults = nextDatabank.load(data);
    if (typeof console != "undefined" && console.debug) {
        // console.debug("about to debug annotationResults");
        // console.debug(annotationResults);
        console.debug("About to create nextQuery");
    }

    var nextQuery = $.rdf({
        databank : nextDatabank
    }).
    // Find an annotation
    where(' ?annotationUri rdf:type tagging:Tagging .').
    // Get label
    optional('?annotationUri tagging:name ?label').
    // Get tagged resource
    optional('?annotationUri tagging:taggedResource ?taggedResource').
    // Get tagger
    optional('?annotationUri foaf:maker ?author').
    // Get ontology term
    optional('?annotationUri moat:tagMeaning ?ontologyTermUri').
    // Get ontology for ontology term
    optional('?ontologyTermUri rdfs:isDefinedBy ?ontologyUri');

    if (typeof console != "undefined" && console.debug) {
        // console.debug("About to debug nextQuery");
        // console.debug(nextQuery);
    }
    var bindings = nextQuery.select();

    if (typeof console != "undefined" && console.debug) {
        console.debug("About to return bindings");
        console.debug(bindings);
    }

    var nodeChildren = [];
    $.each(bindings, function(index, value) {
        var nextChild = {};
        nextChild.annotationUri = value.annotationUri.value;

        if (value.label) {
            nextChild.label = value.label.value;
        }
        else {
            nextChild.label = value.annotationUri.value;
        }

        if (value.taggedResource) {
            nextChild.taggedResource = value.taggedResource.value;
        }
        else {
            nextChild.taggedResources = "";
        }

        if (value.author) {
            nextChild.author = value.author.value;
        }
        else {
            nextChild.author = "";
        }

        if (value.ontologyTermUri) {
            nextChild.ontologyTermUri = value.ontologyTermUri.value;
        }
        else {
            nextChild.ontologyTermUri = "";
        }

        if (value.ontologyUri) {
            nextChild.ontologyUri = value.ontologyUri.value;
        }
        else {
            nextChild.ontologyUri = "";
        }

        nodeChildren.push(nextChild);
    });

    if (typeof console != "undefined" && console.debug) {
        console.debug("About to return nodeChildren");
        console.debug(nodeChildren);
    }

    // Sort annotations
    // TODO: Make this configurable/selectable by users
    nodeChildren.sort(function(a, b) {
        a = a.label.toLowerCase();
        b = b.label.toLowerCase();
        return a > b ? 1 : a < b ? -1 : 0;
    });

    return nodeChildren;
};

/**
 * Parses an RDF/JSON string containing counts and returns the counts as an
 * array of SPARQL bindings.
 */
oas.parseCountsRdf = function(/* string */baseUrl, /* rdf/json */data) {
    // Parser for results from count services, including
    // annotation/countbyontologytermuri
    if (typeof console != "undefined" && console.debug) {
        console.debug("Parsing count results");
        console.debug(baseUrl);
        console.debug(data);
    }

    var nextDatabank = oas.newDatabank();

    var rdfSearchResults = nextDatabank.load(data);
    if (typeof console != "undefined" && console.debug) {
        // console.debug("about to debug rdfSearchResults");
        // console.debug(rdfSearchResults);
        console.debug("About to create nextQuery");
    }

    var nextQuery = $.rdf({
        databank : nextDatabank
    }).where(' ?result oas:uriDirectCount ?count');

    var bindings = nextQuery.select();

    return bindings;
};

/**
 * Returns the first integer count from the given SPARQL bindings. Returns 0 if
 * no counts were found.
 */
oas.getFirstIntegerCount = function(/* object */bindings) {
    var result = 0;
    $.each(bindings, function(index, value) {
        if (value.count) {
            result = value.count.value;
        }
    });
    return result;
};

/**
 * Annotation URI identifier that returns an RDFa URI matching the target of the
 * annotation handler based on its "about" attribute.
 */
oas.defaultRdfaIdentifier = function(/* object */target) {
    return target.attributes.getNamedItem("about").value;
};

oas.defaultRdfaFinder = function() {
    return $("[about]");
};

/**
 * Creates a default annotation dialog.
 */
oas.defaultDialog = function(/* object */resultDialog) {
    var div = $("<div>");
    var form = $("<form>");

    var annotationTargetInput = $("<input>", {
        type : 'text'
    });
    annotationTargetInput.addClass("oasAnnotationTarget");

    var annotationTargetLabel = $("<label>").text("Annotation target:").append(annotationTargetInput);
    form.append(annotationTargetLabel);

    var annotationContentInput = $("<input>", {
        type : 'text'
    });
    var annotationContentLabel = $("<label>").text("Annotation content:").append(annotationContentInput);
    form.append(annotationContentLabel);

    var ontologyInput = $("<input>", {
        type : 'text'
    });
    var ontologyUriInput = $("<input>", {
        type : 'hidden'
    });
    var ontologyLabel = $("<label>").text("Ontology:").append(ontologyInput).append(ontologyUriInput);
    form.append(ontologyLabel);

    var ontologyTermInput = $("<input>", {
        type : 'text'
    });
    var ontologyTermUriInput = $("<input>", {
        type : 'hidden'
    });
    var ontologyTermLabel = $("<label>").text("Ontology Term:").append(ontologyTermInput).append(ontologyTermUriInput);
    form.append(ontologyTermLabel);

    // <div id="dialog" title="Tab data">
    // <form id="ontology_annotation_form">
    // <label id="label-annotationtarget" for="annotation_target">Annotation
    // target :</label>
    // <input id="annotation_target" size="150" name="annotation_target"
    // type="text">&nbsp;</input>
    //        
    // <label id="label-annotationcontent"
    // for="annotation_content">Content</label>
    // <textarea name="annotation_content" id="annotation_content"
    // class="ui-widget-content ui-corner-all"></textarea>
    //        
    // <label id="label-ontology" for="ontology">Ontology</label>
    // <img id="ontology-icon"
    // src="${baseUrl}resources/static/images/transparent_1x1.png"></img>
    // <input name="ontology" id="ontology" type="text">&nbsp;</input>
    // <input name="ontology-id" type="text" id="ontology-id"
    // value="">&nbsp;</input>
    // <p id="ontology-description">&nbsp;</p>
    //        
    // <label id="label-ontologytermlabel" for="ontologytermlabel">Ontology
    // term</label>
    // <input name="ontologytermlabel" id="ontologytermlabel"
    // type="text">&nbsp;</input>
    // <input name="ontologytermuri" id="ontologytermuri"
    // type="text">&nbsp;</input>
    // </form>
    // </div>

    div.append(form);
    $('body').append(div);

    var dialog = div
            .dialog({
                autoOpen : false,
                title : 'Create annotation dialog',
                resizable : true,
                width : 550,
                buttons : [
                        {
                            text : "Create annotation",
                            click : function(eventObject) {
                                var result = oas.defaultDialogSubmit(annotationTargetInput, annotationContentInput,
                                        ontologyInput, ontologyUriInput, ontologyTermInput, ontologyTermUriInput,
                                        resultDialog);
                                if (!result) {
                                    // Keep the dialog open so they can try to
                                    // fix it
                                    eventObject.preventDefault();
                                    return false;
                                }
                                else {
                                    $(this).dialog("close");
                                }
                            }
                        }, {
                            text : "Cancel",
                            click : function() {
                                $(this).dialog("close");
                            }
                        } ]
            });

    // link in the autocomplete code after the elements are in the DOM so that
    // the correct handler behaviour is not contraindicated
    oas.constructAutocomplete(ontologyInput, ontologyUriInput, ontologyTermInput, ontologyTermUriInput, dialog);

    return dialog;
};

/**
 * 
 * @returns A Dialog object that has been initialised but not opened. Must
 *          contain an element with the class "oasAnnotationResult" that will be
 *          populated with the details of the annotation, including a link.
 */
oas.defaultResultDialog = function() {
    var div = $("<div>").val("Annotation created successfully");
    var annotationResultLink = $("<div>").addClass("oasAnnotationResult");
    div.append(annotationResultLink);
    $('body').append(div);

    var resultDialog = div.dialog({
        autoOpen : false,
        title : 'Annotation created',
        resizable : true,
        width : 550,
        buttons : [ {
            text : "Close",
            click : function() {
                $(this).dialog("close");
            }
        } ]
    });

    return resultDialog;
};

/**
 * Creates a default login dialog.
 */
oas.defaultLoginDialog = function() {
    var div = $("<div>");
    var form = $("<form>");
    div.append(form);
    form.text("You must login to Ontology Annotation Services before creating annotations.");

    $('body').append(div);

    return div.dialog({
        autoOpen : false,
        title : 'Login to Ontology Annotation Services',
        resizable : true,
        width : 550,
        buttons : [ {
            text : "Login",
            click : function() {
                window.location.href = oas.baseUrl + "login";
            }
        }, {
            text : "Cancel",
            click : function() {
                $(this).dialog("close");
            }
        } ]
    });
};

/**
 * Pushes the annotated object URI into the dialog.
 */
oas.defaultAnnotationTargetHandler = function(eventObject, dialog, annotatedObjectURI) {
    // link in this value to the dialog
    return dialog.find(".oasAnnotationTarget").val(annotatedObjectURI);
};

/**
 * Setup an annotation button object. The click handler is added to the object
 * that is returned. To add the click handler to the context object, return the
 * contextObject from the function.
 * 
 * @param contextObject
 *            A JQuery object wrapping the context for one of the objects
 *            returned by annotationTargetFinder that is known to have an
 *            identfiable object attached to it somehow.
 */
oas.defaultButtonPrepend = function(contextObject) {
    var div = $("<div>");
    div.css("float", "right");

    var button = $("<button>").text("Annotate me!");

    // Add the button as the first child
    contextObject.prepend(div);

    div.append(button);

    return div;
};

/**
 * Setup an annotation button object. The click handler is added to the object
 * that is returned. To add the click handler to the context object, return the
 * contextObject from the function.
 * 
 * @param contextObject
 *            A JQuery object wrapping the context for one of the objects
 *            returned by annotationTargetFinder that is known to have an
 *            identfiable object attached to it somehow.
 */
oas.buttonAppend = function(contextObject) {
    var div = $("<div>");
    div.css("float", "right");

    var button = $("<button>").text("Annotate me!");

    // Add the button as the first child
    contextObject.append(div);

    div.append(button);

    return div;
};

/**
 * Adds the annotation count for the given URI positioning it based on the given
 * contextObject.
 * 
 * @param contextObject
 * @param nextURI
 * @param count
 */
oas.defaultAddAnnotationCountPrepend = function(contextObject, nextURI, count) {
    var div = $("<div>");
    if (typeof count != "undefined" && count > 0) {
        var span = $('<span>').text("Current annotations: " + count);
        var a = $('<a>').attr('href', oas.baseUrl + "annotation/byobjecturi?objecturi=" + nextURI);
        div.css("float", "right");
        
        a.append(span);
        div.append(a);
    }
    contextObject.prepend(div);
    
    return div;
};

/**
 * Adds the annotation count for the given URI positioning it based on the given
 * contextObject.
 * 
 * @param contextObject
 * @param nextURI
 * @param count
 */
oas.addAnnotationCountAppend = function(contextObject, nextURI, count) {
    var div = $("<div>");
    if (typeof count != "undefined" && count > 0) {
        var span = $('<span>').text("Current annotations: " + count);
        var a = $('<a>').attr('href', oas.baseUrl + "annotation/byobjecturi?objecturi=" + nextURI);
        div.css("float", "right");
        
        a.append(span);
        div.append(a);
    }
    contextObject.append(div);
    
    return div;
};

/**
 * The default annotation options, including handlers for
 * finding/identifying/handling annotation targets, along with an annotation
 * creation dialog and a login dialog if the user is not logged in when they
 * attempt to create an annotation.
 */
oas.defaultAnnotationOptions = {
    annotationTargetFinder : oas.defaultRdfaFinder,
    annotationTargetIdentifier : oas.defaultRdfaIdentifier,
    annotationTargetHandler : oas.defaultAnnotationTargetHandler,
    setupAnnotationButton : oas.defaultButtonPrepend,
    addAnnotationCount : oas.defaultAddAnnotationCountPrepend,
    dialog : oas.defaultDialog,
    loginDialog : oas.defaultLoginDialog,
    resultDialog : oas.defaultResultDialog
};

/**
 * Adds annotation handlers to all nodes matching the given path, and when they
 * are clicked on, the dialog is opened so that they can fill in the details for
 * their annotation.
 * 
 * Default settings are taken from oas.defaultAnnotationOptions
 */
oas.addAnnotationHandlers = function(/* object */options) {
    var settings = $.extend({}, oas.defaultAnnotationOptions, options);

    settings.annotationTargetFinder().each(function(index, value) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("found an annotatable object:");
            console.debug(value);
        }
        var nextURI = settings.annotationTargetIdentifier(value);

        // If the annotation target identifier did not produce a value
        // then skip this element
        if (nextURI) {
            var context = $(this);
            // If there was a URI identified for this event then setup a button
            // that they can use to annotate this target
            var div = settings.setupAnnotationButton(context);
            oas.annotationCountByTarget(nextURI, true, function(count) {
                settings.addAnnotationCount(context, nextURI, count);
            });

            div.click(function(eventObject) {
                // eventObject.preventDefault();

                if (typeof console != "undefined" && console.debug) {
                    console.debug("click event fired:");
                    // console.debug(eventObject);
                }

                // Create all of the necessary dialogs, but keep them hidden
                // from view at this point
                var dialog = settings.dialog(settings.resultDialog());
                var loginDialog = settings.loginDialog();

                // Find the target based on this event, and insert the
                // predefined URI for this click handler into the dialog which
                // was just created
                settings.annotationTargetHandler(eventObject, dialog, nextURI);

                // Check authentication and either open the annotation dialog if
                // the user is authenticated, or show the dialog offering to
                // send them to the login page otherwise
                oas.checkAuthentication(oas.openAnnotationDialog, dialog, oas.openLoginDialog, loginDialog);

                // further prevent the default action, e.g.,
                // following a link
                return false;
            });
        }
        else {
            if (typeof console != "undefined" && console.debug) {
                console.debug("Could not identify URI for annotatable object:");
                console.debug(value);
                console.debug(nextURI);
            }
        }
    });
};

/**
 * Integrates a check for authentication with callbacks using either the success
 * callback with the success dialog or the failurecallback with the failure
 * dialog as appropriate.
 */
oas.checkAuthentication = function(/* function */successCallback, /* object */successDialog,
/* function */failureCallback, /* object */failureDialog) {
    $.ajax({
        url : oas.baseUrl + "whoami",
        dataType : "text",
        success : function(data, status) {
            if (typeof console != "undefined" && console.debug) {
                console.debug("Was authenticated");
                console.debug(data);
            }
            successCallback(successDialog);
        },
        error : function(xhr, ajaxOptions, thrownError) {
            if (typeof console != "undefined" && console.debug) {
                console.debug("Was not authenticated");
            }
            failureCallback(failureDialog);
            // TODO: Only do redirect if status was 401
            // window.location.href = oas.baseUrl + "login";
        }
    });
};

oas.openAnnotationDialog = function(/* object */dialog) {
    dialog.dialog('open');

    if (typeof console != "undefined" && console.debug) {
        console.debug("Autocomplete dialog open");
    }
};

oas.openLoginDialog = function(/* object */dialog) {
    dialog.dialog('open');

    if (typeof console != "undefined" && console.debug) {
        console.debug("Login dialog open");
    }
};

/**
 * Construct an autocomplete handler for ontology labels. When the autocomplete
 * handler runs and a user selects an ontology it will setup the handler for the
 * ontology term search autocomplete input.
 */
oas.constructAutocomplete = function(/* object */ontologyLabel, /* object */ontologyUri, /* object */
ontologyTermLabel, /* object */ontologyTermUri, dialog) {
    // used to prevent race conditions between autocomplete requests
    var requestIndex = 0;

    ontologyLabel.autocomplete({
        serviceUrl : oas.baseUrl + 'ontologymanager/labelsearch?searchterm=',
        minChars : 3,
        width : 300,
        // delimiter : /(,|;)\s*/,
        deferRequestBy : 300, // miliseconds
        focus : function(event, ui) {
            ontologyLabel.val(ui.item.label);
            // $( "#ontology-icon" ).attr( "src",
            // oas.baseUrl+"resources/static/images/" + ui.item.icon );
            return false;
        },
        source : function(request, response) {
            if (typeof console != "undefined" && console.debug) {
                // console.debug("in source function");
                // console.debug(this.options.serviceUrl);
                // console.debug(request);
            }
            if (self.xhr) {
                self.xhr.abort();
            }
            self.xhr = $.ajax({
                url : this.options.serviceUrl + request.term,
                dataType : "json",
                autocompleteRequest : ++requestIndex,
                success : function(data, status) {
                    if (this.autocompleteRequest === requestIndex) {
                        response(oas.parseSearchResults(this.url, data));
                    }
                },
                error : function(xhr, ajaxOptions, thrownError) {
                    if (this.autocompleteRequest === requestIndex) {
                        response([]);
                    }
                }
            });
        },
        select : function(event, ui) {
            if (typeof console != "undefined" && console.debug) {
                console.debug("in select function");
                console.debug(event);
                console.debug(ui);
                console.debug(ontologyLabel);
                console.debug(ontologyUri);
            }
            ontologyLabel.val(ui.item.label);
            ontologyUri.val(ui.item.value.toString());
            // $( "#ontology-description" ).html( ui.item.desc );
            // $( "#ontology-icon" ).attr( "src", "images/" + ui.item.icon );
            return false;
        }
    }).data("ui-autocomplete")._renderItem = function(ul, item) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("in _renderItem function");
            console.debug(ul);
            console.debug(item);
            console.debug(ontologyLabel);
            console.debug(ontologyUri);
        }
        return $("<li></li>").data("item.autocomplete", item).append("<a>" + item.label + "</a>").appendTo(ul);
        // TODO: Renable desc when it is available
        // return $("<li></li>").data("item.autocomplete", item)
        // .append("<a>" + item.label + "<br />" + item.desc +
        // "</a>").appendTo(ul);
    };

    ontologyTermLabel.autocomplete({
        serviceUrl : oas.baseUrl + 'ontologymanager/search?searchterm=',
        minChars : 3,
        width : 300,
        delimiter : /(,|;)\s*/,
        deferRequestBy : 500, // miliseconds
        source : function(request, response) {
            if (self.xhr) {
                self.xhr.abort();
            }

            var requestUrl = this.options.serviceUrl + request.term;

            var nextUri = $.trim(ontologyUri.val());
            if (nextUri.length > 0) {
                requestUrl = requestUrl + "&ontologyUri=" + nextUri;
            }

            self.xhr = $.ajax({
                url : requestUrl,
                dataType : "json",
                autocompleteRequest : ++requestIndex,
                success : function(data, status) {
                    if (this.autocompleteRequest === requestIndex) {
                        response(oas.parseSearchResults(this.url, data));
                    }
                },
                error : function(xhr, ajaxOptions, thrownError) {
                    if (this.autocompleteRequest === requestIndex) {
                        response([]);
                    }
                }
            });
        },
        select : function(event, ui) {
            if (typeof console != "undefined" && console.debug) {
                console.debug("Option selected");
                console.debug(event);
                console.debug(ui);
            }
            ontologyTermLabel.val(ui.item.label);
            ontologyTermUri.val(ui.item.value.toString());
            return false;
        }
    });

    // These are necessary to keep the Z-order correct, or else the autocomplete
    // widget stays behind the dialog and is not visible
    ontologyLabel.autocomplete("widget").insertAfter(dialog.parent());
    ontologyTermLabel.autocomplete("widget").insertAfter(dialog.parent());

};

/**
 * Creates an ontology browser in the given path set to work with the given
 * ontology URI and optional version URI.
 * 
 * TODO: Refactor this to use a settings object with defaults
 */
oas.ontology.createBrowser = function(/* string */treePath,/* string */ontologyUri, /* string */ontologyVersionUri, /* function */
onActivateHandler, /* function */onDeactivateHandler, /* function */onFocusHandler, /* function */
onBlurHandler, /* string */btnSortStrategyPath, /* object */rootNodeData) {
    // Initialize the tree inside the <div>element.
    $(treePath).dynatree(
            {
                title : "OAS Ontology Browser",
                onActivate : onActivateHandler,
                onDeactivate : onDeactivateHandler,
                onFocus : onFocusHandler,
                onBlur : onBlurHandler,
                onLazyRead : function(node) {
                    if (typeof console != "undefined" && console.debug) {
                        console.debug("onLazyRead");
                    }
                    var nextUrl = "";
                    if (ontologyVersionUri && ontologyVersionUri !== "") {
                        nextUrl = oas.baseUrl + "ontologybrowser/?ontologyUri=" + encodeURIComponent(ontologyUri)
                                + "&ontologyVersionUri=" + encodeURIComponent(ontologyVersionUri);
                    }
                    else {
                        nextUrl = oas.baseUrl + "ontologybrowser/?ontologyUri=" + encodeURIComponent(ontologyUri);
                    }

                    // if the node had a URL attached, use it instead of the
                    // default
                    if (node && node.data && node.data.url) {
                        if (typeof console != "undefined" && console.debug) {
                            console.debug("onLazyRead, node.data.url=" + node.data.url);
                        }
                        nextUrl = node.data.url;
                    }
                    $.ajax({
                        url : nextUrl,
                        beforeSend : function(xhr) {
                            xhr.setRequestHeader("Accept", "application/rdf+json");
                        }
                    }).done(
                            function(data) {
                                var nextDatabank = oas.newDatabank();

                                if (typeof console != "undefined" && console.debug) {
                                    // console.debug("nextDatabank");
                                    // console.debug(nextDatabank);
                                }

                                var rdfSearchResults = nextDatabank.load(data);

                                if (typeof console != "undefined" && console.debug) {
                                    // console.debug("about to debug
                                    // rdfSearchResults");
                                    // console.debug(rdfSearchResults);
                                    // console.debug("About to create labelQuery
                                    // updated");
                                }

                                var labelQuery = $.rdf({
                                    databank : nextDatabank
                                }).where('?url rdfs:label ?title')
                                        .optional('?url oas:uriDirectCount ?directChildCount').optional(
                                                '?url oas:uriRecursiveCount ?recursiveChildCount');

                                if (typeof console != "undefined" && console.debug) {
                                    // console.debug("About to debug
                                    // labelQuery");
                                    // console.debug(labelQuery);
                                }

                                var bindings = labelQuery.select();

                                if (typeof console != "undefined" && console.debug) {
                                    // console.debug("About to return
                                    // bindings");
                                    // console.debug(bindings);
                                }
                                var nodeChildren = [];
                                $.each(bindings, function(index, value) {
                                    var nextChild = {};
                                    // Setup defaults
                                    nextChild.tooltip = value.url.value.toString() + "";
                                    nextChild.ontologyTermUri = value.url.value.toString();
                                    nextChild.directChildCount = 0;
                                    nextChild.recursiveChildCount = 0;

                                    // Add counts
                                    if (value.directChildCount) {
                                        nextChild.directChildCount = value.directChildCount.value;
                                        nextChild.tooltip = value.directChildCount.value + "";
                                    }
                                    if (value.recursiveChildCount) {
                                        nextChild.recursiveChildCount = value.recursiveChildCount.value;
                                        nextChild.tooltip = value.recursiveChildCount.value + "";
                                    }

                                    if (value.title) {
                                        nextChild.title = value.title.value + " (" + nextChild.directChildCount
                                                + " => " + nextChild.recursiveChildCount + ")";
                                    }
                                    else {
                                        if (value.url.value.toString() == "http://www.w3.org/2002/07/owl#Thing") {
                                            nextChild.title = "OWL Thing (" + nextChild.directChildCount + " => "
                                                    + nextChild.recursiveChildCount + ")";
                                        }
                                        else {
                                            nextChild.title = value.url.value + " (" + nextChild.directChildCount
                                                    + " => " + nextChild.recursiveChildCount + ")";
                                        }
                                    }

                                    // if (typeof console != "undefined" &&
                                    // console.debug) {
                                    // console.debug(nextChild["tooltip"]);
                                    // }

                                    if (ontologyVersionUri && ontologyVersionUri !== "") {
                                        nextChild.url = oas.baseUrl + "ontologybrowser/?ontologyUri="
                                                + encodeURIComponent(ontologyUri) + "&ontologyVersionUri="
                                                + encodeURIComponent(ontologyVersionUri) + "&parentUri="
                                                + encodeURIComponent(value.url.value.toString());
                                    }
                                    else {
                                        nextChild.url = nextUrl = oas.baseUrl + "ontologybrowser/?ontologyUri="
                                                + encodeURIComponent(ontologyUri) + "&parentUri="
                                                + encodeURIComponent(value.url.value.toString());
                                    }

                                    // if there is a known count, only configure
                                    // lazy loading for children that report
                                    // more than 0 children
                                    if (value.directChildCount) {
                                        if (value.directChildCount.value > 0) {
                                            nextChild.isLazy = true;
                                        }
                                        else {
                                            nextChild.isLazy = false;
                                        }
                                    }
                                    else if (value.recursiveChildCount) {
                                        if (value.recursiveChildCount.value > 0) {
                                            nextChild.isLazy = true;
                                        }
                                        else {
                                            nextChild.isLazy = false;
                                        }
                                    }
                                    else {
                                        // Fall back to all children being lazy
                                        // if we don't know the count
                                        nextChild.isLazy = true;
                                    }

                                    nodeChildren.push(nextChild);
                                });
                                node.addChild(nodeChildren);

                                node.sortChildren(oas.ontology.currentSorter, false);

                                // Remove the 'loading...' status:
                                node.setLazyNodeStatus(DTNodeStatus_Ok);
                                return false;
                            }).fail(function(jqXHR, textStatus, errorThrown) {
                        if (typeof console != "undefined" && console.debug) {
                            console.debug("Error during lazy loading of node:");
                            console.debug(jqXHR);
                            console.debug(textStatus);
                            console.debug(errorThrown);
                        }
                        node.setLazyNodeStatus(DTNodeStatus_Error);
                        return false;
                    });

                    return false;
                }
            });

    /**
     * Detect changes in the sort radio buttons and change the current sorter to
     * match.
     */
    $(btnSortStrategyPath).change(function() {
        var node = $(treePath).dynatree("getRoot");
        var sortStrategy = $(btnSortStrategyPath + ":checked").val();
        if (sortStrategy === "strategyLexical") {
            oas.ontology.currentSorter = oas.ontology.sortLexical;
        }
        else if (sortStrategy === "strategyDirectChildren") {
            oas.ontology.currentSorter = oas.ontology.sortDirectChildCount;
        }
        else if (sortStrategy === "strategyAllChildren") {
            oas.ontology.currentSorter = oas.ontology.sortAllChildCount;
        }
        node.sortChildren(oas.ontology.currentSorter, true);
    });

    // TODO: What was the following meant to do?
    var rootNode = $(treePath).dynatree("getRoot");
    var nodeList = rootNode.getChildren();
    var currentRootNodeData = nodeList[0].data;
    $.each(rootNodeData, function(index, value) {
        currentRootNodeData[index] = value;
    });
};

/**
 * Finds the number of annotations linked to the given ontology term, optionally
 * specifically linked to the given ontology and version.
 * 
 * The callback is called with either the "undefined" object if the query
 * failed, or an integer if the query succeeded.
 * 
 * @param ontologyTermUri
 * @param ontologyUri
 * @param ontologyVersionUri
 * @param recursive
 * @param callback
 */
oas.annotationCount = function(ontologyTermUri, ontologyUri, ontologyVersionUri, recursive, callback) {
    var queryargs = "ontologyTermUri=" + encodeURIComponent(ontologyTermUri);

    if (typeof ontologyUri != "undefined") {
        queryargs = queryargs + "&ontologyUri=" + encodeURIComponent(ontologyUri);
    }

    if (typeof ontologyVersionUri != "undefined") {
        queryargs = queryargs + "&ontologyVersionUri=" + ontologyVersionUri;
    }

    if (recursive) {
        queryargs = queryargs + "&recursive=true";
    }
    else {
        queryargs = queryargs + "&recursive=false";
    }

    $.ajax({
        type : "GET",
        url : oas.baseUrl + "annotation/countbyontologytermuri?" + queryargs,
        beforeSend : function(xhr) {
            xhr.setRequestHeader("Accept", "application/rdf+json");
        }
    }).done(function(data) {
        var bindings = oas.parseCountsRdf(ontologyTermUri, data);
        callback(oas.getFirstIntegerCount(bindings));
    }).fail(function(jqXHR, textStatus) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Get annotation counts by ontology term URI request failed: " + textStatus);
        }
        callback(undefined);
    });
};

/**
 * Fetch the count of annotations linked to the given object/target URI,
 * optionally using recursion, and return the count to the callback.
 * 
 * @param nextURI
 * @param recursive
 * @param callback
 */
oas.annotationCountByTarget = function(nextURI, recursive, callback) {
    $.ajax({
        type : "GET",
        url : oas.baseUrl + "annotation/countbyobjecturi?objectUri=" + nextURI,
        beforeSend : function(xhr) {
            xhr.setRequestHeader("Accept", "application/rdf+json");
        }
    }).done(function(data) {
        var bindings = oas.parseCountsRdf(nextURI, data);
        callback(oas.getFirstIntegerCount(bindings));
    }).fail(function(jqXHR, textStatus) {
        if (typeof console != "undefined" && console.debug) {
            console.debug("Get annotation counts by object URI request failed: " + textStatus);
        }
        callback(undefined);
    });
};

/**
 * Loads details for ontology item into the details tab.
 */
oas.ontology.loadDetails = function(ontologyTermUri, ontologyUri, ontologyVersionUri, directAnnotationCountPath,
        allAnnotationCountPath, annotationListPath) {

    // Get recursive count
    oas.annotationCount(ontologyTermUri, ontologyUri, ontologyVersionUri, true, function(result) {
        if (typeof result != "undefined") {
            $(allAnnotationCountPath).text(result);
        }
        else {
            $(allAnnotationCountPath).text("-");
        }
    });

    // Get non-recursive count
    oas.annotationCount(ontologyTermUri, ontologyUri, ontologyVersionUri, false, function(result) {
        if (typeof result != "undefined") {
            $(directAnnotationCountPath).text(result);
        }
        else {
            $(directAnnotationCountPath).text("-");
        }
    });

    var queryargs = "ontologyTermUri=" + encodeURIComponent(ontologyTermUri) + "&ontologyUri=" + ontologyUri;

    if (ontologyVersionUri !== "") {
        queryargs = queryargs + "&ontologyVersionUri=" + ontologyVersionUri;
    }

    $.ajax({
        type : "GET",
        url : oas.baseUrl + "annotation/byontologytermuri?" + queryargs,
        beforeSend : function(xhr) {
            xhr.setRequestHeader("Accept", "application/rdf+json");
        }
    }).done(function(data) {
        $(annotationListPath).empty();
        var bindings = oas.parseAnnotationsResults(ontologyTermUri, data);
        if (typeof console != "undefined" && console.debug) {
            console.debug("Get annotations by ontology term URI successful");
            console.debug(bindings);
        }
        $.each(bindings, function(index, value) {
            $(annotationListPath).append(
            // annotation list item
            $('<li>').append(
            // link
            $('<a>').attr('href', oas.baseUrl + "annotation/id/?annotationid=" + value.annotationUri).append(
            // label
            $('<span>').text(value.label + " by " + value.author))));
        });
    }).fail(function(jqXHR, textStatus) {
        $(annotationListPath).empty();
        if (typeof console != "undefined" && console.debug) {
            console.debug("Get annotations by ontology term URI failed: " + textStatus);
        }
    });

};

/**
 * Ontology browser tree sorting implementations.
 */

/**
 * Sort ontology browser tree by alphabet/lexical values.
 */
oas.ontology.sortLexical = function(a, b) {
    a = a.data.title.toLowerCase();
    b = b.data.title.toLowerCase();
    return a > b ? 1 : a < b ? -1 : 0;
};

/**
 * Sort ontology browser tree by direct children count.
 */
oas.ontology.sortDirectChildCount = function(a, b) {
    var aCount = a.data.directChildCount;
    var bCount = b.data.directChildCount;

    if (aCount < bCount) {
        return 1;
    }

    if (aCount > bCount) {
        return -1;
    }

    // Sort secondary by all child count before sorting lexically
    aCount = a.data.recursiveChildCount;
    bCount = b.data.recursiveChildCount;

    if (aCount < bCount) {
        return 1;
    }

    if (aCount > bCount) {
        return -1;
    }

    // If the counts are the same revert to sorting by lexical title
    // value, case-insensitive
    aTitle = a.data.title.toLowerCase();
    bTitle = b.data.title.toLowerCase();

    return aTitle > bTitle ? 1 : aTitle < bTitle ? -1 : 0;
};

/**
 * Sort ontology browser tree by total child count.
 */
oas.ontology.sortAllChildCount = function(a, b) {
    var aCount = a.data.recursiveChildCount;
    var bCount = b.data.recursiveChildCount;

    if (aCount < bCount) {
        return 1;
    }

    if (aCount > bCount) {
        return -1;
    }

    // Sort secondary by direct child count before sorting lexically
    aCount = a.data.directChildCount;
    bCount = b.data.directChildCount;

    if (aCount < bCount) {
        return 1;
    }

    if (aCount > bCount) {
        return -1;
    }

    // If the counts are the same revert to sorting by lexical title
    // value, case-insensitive
    aTitle = a.data.title.toLowerCase();
    bTitle = b.data.title.toLowerCase();

    return aTitle > bTitle ? 1 : aTitle < bTitle ? -1 : 0;
};

/**
 * Defines the default sorter for the ontology browser tree.
 */
oas.ontology.currentSorter = oas.ontology.sortAllChildCount;
